const axios = require('axios');
const API = require('../config/API');
const querystring = require('querystring');
const io = require('socket.io-client');
const commonMethods = require('./common-methods');
const constants = require('./constants');
const SocketClient = require('./websocket-client');
const utils = require('./utils');
const logger = require('./logger');

class B2CAPI {
    constructor(configParams) {
        const self = this;
        self.url = configParams.url;
        self.loginType = API.loginType;
        self.secondAuthType = API.secondAuthType;
        self.apiKey = configParams.apiKey;
        self.x_api_key = configParams.x_api_key;
        self.source = API.source;
        self.logonResponse = null;
        self.ioClient = null;
        self.objSocket = new SocketClient();
        if (self.debug) self.objSocket.debug = true; 
        self.eventEmitter = commonMethods.pubsub();

        const requestInstance = axios.create({
            baseURL: self.url,
            timeout: 3000
        });

        requestInstance.interceptors.request.use(function (request) {
            if (self.debug) {
                logger.info(`Request: ${JSON.stringify({
                    baseURL: request.baseURL,
                    url: request.url,
                    method: request.method,
                    headers: request.headers,
                    data: request.data
                })}`);
            };
            return request;
        });

        requestInstance.interceptors.response.use(function (response) {
            if (self.debug) {
                logger.info(`Response: ${JSON.stringify({
                    status: response.status,
                    statusText: response.statusText,
                    headers: response.headers,
                    data: response.data
                })}`);
            }

            if (response.status === 200) {
                if (response.data.status.toLowerCase() == 'success') {
                    return response.data;
                };
            };
            return response;
        }, (error) => {
            if (error.response) {
                if (self.debug) logger.error(`Response: ${JSON.stringify(error.response.data)}`);
                throw error.response.data;
            };
            if (self.debug) logger.error(`Response: ${JSON.stringify(error)}`);
            throw error;
        });

        function request(method, url, pathVar, queryParams, body, headers) {
            let URL = (method === 'GET' || method === 'DELETE') && queryParams ? `${url}?${querystring.stringify(queryParams)}` : url;
            if (URL.indexOf('{') !== -1) {
                for (const key in pathVar) {
                    if (pathVar.hasOwnProperty(key)) {
                        URL = URL.replace(`{${key}}`, pathVar[key]);
                    };
                };
            };
            const options = {
                method,
                url: URL,
                data: JSON.stringify(body),
                headers: {
                    'Content-Type': 'application/json',
                    ...headers
                },
            };
            if (self.logonResponse) {
                options['headers']['Authorization'] = `Bearer ${self.logonResponse.access_token}`;
            };
            if (self.x_api_key) {
                options['headers']['x-api-key'] = `${self.x_api_key}`;
            };

            return requestInstance.request(options);
        };

        self.login = async function (paramsObj) {
            try {
                const body = {
                    user_id: paramsObj.userId,
                    login_type: self.loginType,
                    password: paramsObj.password,
                    second_auth_type: self.secondAuthType,
                    second_auth: paramsObj.totp,
                    api_key: self.apiKey,
                    source: self.source
                };
                const data = await request('POST', API.session, null, null, body);
                if (data && data.data) {
                    self.logonResponse = data.data;
                    return data;
                };
                return data;
            } catch (error) {
                if (self.debug) logger.error(`Login ${JSON.stringify(error)}`);
                throw error;
            };
        };

        self.balance = function () {
            return request('GET', API.balance);
        };

        self.validateSession = function () {
            return request('PUT', API.session);
        };

        self.logout = function () {
            return request('DELETE', API.session);
        };

        self.getOrderBook = function (params) {
            let queryParams = {
                offset: '1',
                limit: '20',
                orderStatus: null,
                order_id: null
            };
            if (params)
                queryParams = { ...queryParams, ...params };
            return request('GET', API.orders, null, queryParams);
        };

        self.getTradeBook = function (params) {
            let queryParams = {
                offset: '1',
                limit: '100'
            };
            if (params)
                queryParams = { ...queryParams, ...params };
            return request('GET', API.trades, null, queryParams);
        };

        self.getOrderHistory = function (params) {
            const pathVar = {
                order_id: params.orderId,
            };
            return request('GET', API.orderHistory, pathVar);
        };

        self.placeOrder = function (params) {
            const body = params;
            return request('POST', API.placeOrder, null, null, body);
        };

        self.modifyOrder = function (params) {
            const pathVar = {
                exchange: params.exchange,
                order_id: params.order_id,
            };
            const body = params;
            return request('PUT', API.modifyOrders, pathVar, null, body);
        };

        self.cancelOrder = function (params) {
            const pathVar = {
                exchange: params.exchange,
                order_id: params.order_id,
            };
            return request('DELETE', API.modifyOrders, pathVar);
        };

        self.placeCoverOrder = function (params) {
            return request('POST', API.coverOrder, null, null, params);
        };

        self.modifyCoverOrder = function (params) {
            const pathVar = {
                exchange: params.exchange,
                order_id: params.order_id,
            };
            const body = params;
            return request('PUT', API.modifyCoverOrder, pathVar, null, body);
        };

        self.cancelCoverOrder = function (params) {
            const pathVar = {
                exchange: params.exchange,
                order_id: params.order_id,
            };
            return request('DELETE', API.modifyCoverOrder, pathVar);
        };

        self.placeBracketOrder = function (params) {
            return request('POST', API.bracketOrder, null, null, params);
        };

        self.modifyBracketOrder = function (params) {
            const pathVar = {
                exchange: params.exchange,
                order_id: params.order_id,
            };
            const body = params;
            return request('PUT', API.modifyBracketOrder, pathVar, null, body);
        };

        self.deleteBracketOrder = function (params) {
            const pathVar = {
                exchange: params.exchange,
                order_id: params.order_id,
            };
            return request('DELETE', API.modifyCoverOrder, pathVar);
        };

        self.placeMultilegOrder = function (params) {
            return request('POST', API.placeMultilegOrder, null, null, params);
        };

        self.cancelMultilegOrder = function (params) {
            const pathVar = {
                order_flag: params.order_flag,
                gateway_order_no: params.gateway_order_no,
            };
            const body = params;
            return request('PUT', API.cancelMultilegOrder, pathVar, null, body);
        };

        self.getPositions = function (params) {
            const pathVar = {
                type: params.type,
            };
            return request('GET', API.positions, pathVar);
        };

        self.positionConvertion = function (params) {
            const body = params;
            return request('PUT', API.positionConvertion, null, null, body);
        };

        self.getHoldings = function () {
            return request('GET', API.holdings);
        };


        // socket.io message socket
        self.messageSocket = {};
        self.messageSocket.connect = function () {
            self.ioClient = io.connect(self.logonResponse.others.messageSocket, { transports: ['websocket'] });
            if (self.debug) logger.info('Connecting message websocket.');
            self.ioClient.on('connect', () => {
                if (self.debug) logger.info('Message websocket: connected.');
                self.ioClient.emit('loginAPI', { 'jToken': self.logonResponse.access_token });
            });
        };

        self.messageSocket.on = function (event, callback) {
            if (self.ioClient) {
                switch (event) {
                    case 'MSG:DATA':
                        self.ioClient.on('MSG:DATA', (response) => {
                            if (response.MessageType) {
                                let mappedResp = utils.mappedMsgSocResp(response);
                                if (self.debug) logger.info(`Message socket: MSG:DATA: ${JSON.stringify(mappedResp)}`);
                                callback(mappedResp);
                            } else {
                                if (self.debug) logger.info(`Message socket: MSG:DATA: ${JSON.stringify(response)}`);
                                callback(response);
                            }
                        });
                        break;
                    case 'disconnect':
                        self.ioClient.on('disconnect', (err) => {
                            if (self.debug) logger.error(`Message socket: disconnect: ${err}`);
                            callback(err);
                        });
                        break;
                    case 'connect_error':
                        self.ioClient.on('connect_error', (err) => {
                            if (self.debug) logger.error(`Message socket: connect_error: ${err}`);
                            callback(err);
                        });
                        break;
                    case 'connect_failed':
                        self.ioClient.on('connect_failed', (err) => {
                            if (self.debug) logger.error(`Message socket: connect_failed: ${err}`);
                            callback(err);
                        });
                        break;
                    default:
                        break;
                };
            };
        };

        // broadcast socket

        self.broadcastSocket = {};
        self.broadcastSocket.connect = async function () {
            const reqLogin = { userId: self.logonResponse.user_id, token: self.logonResponse.access_token };
            const socketReq = { SocketUrl: self.logonResponse.others.broadCastSocket, objLogin: reqLogin };
            self.objSocket.initialize(socketReq);
        };

        self.broadcastSocket.touchlineSub = function (lstScrip) {
            if (typeof (lstScrip) != 'object' || lstScrip.length <= 0) {
                return { data: lstScrip, message: 'Invalid request' };
            };
            const reqTL = { arrScrip: lstScrip, operation: 1 };
            self.objSocket.sendTouchlineReq(reqTL);
        };

        self.broadcastSocket.touchlineUnSub = function (lstScrip) {
            if (typeof (lstScrip) != 'object' || lstScrip.length <= 0) {
                return { data: lstScrip, message: 'Invalid request' };
            };
            const reqTL = { arrScrip: lstScrip, operation: 2 };
            self.objSocket.sendTouchlineReq(reqTL);
        };

        self.broadcastSocket.bestFiveSub = function (paramsObj) {
            const reqB5 = { ...paramsObj, operation: 1 };
            self.objSocket.sendBestFiveReq(reqB5);
        };

        self.broadcastSocket.bestFiveUnSub = function (paramsObj) {
            const reqB5 = { ...paramsObj, operation: 2 };
            self.objSocket.sendBestFiveReq(reqB5);
        };

        /**
        * connect listener for event emitter
        *
        */
        self.broadcastSocket.on = function (event, callback) {
            switch (event) {
                case constants.EVT_SCK_OPEN:
                    self.eventEmitter.on(constants.EVT_SCK_OPEN, (data) => {
                        if (self.debug) logger.info(`Broadcast socket: Open: ${JSON.stringify(data)}`);
                        callback(data);
                    });
                    break;
                case constants.EVT_SCK_CLOSE:
                    self.eventEmitter.on(constants.EVT_SCK_CLOSE, (data) => {
                        if (self.debug) logger.info(`Broadcast socket: Close: ${JSON.stringify(data)}`);
                        callback(data);
                    });
                    break;
                case constants.EVT_SCK_LOGIN:
                    self.eventEmitter.on(constants.EVT_SCK_LOGIN, (data) => {
                        if (self.debug) logger.info(`Broadcast socket: Login: ${JSON.stringify(data)}`);
                        callback(data);
                    });
                    break;
                case constants.EVT_SCK_ERROR:
                    self.eventEmitter.on(constants.EVT_SCK_ERROR, (data) => {
                        if (self.debug) logger.info(`Broadcast socket: Error: ${JSON.stringify(data)}`);
                        callback(data);
                    });
                    break;
                case constants.EVT_SCK_BESTFIVE:
                    self.eventEmitter.on(constants.EVT_SCK_MKTDEPTH, (data) => {
                        if (self.debug) logger.info(`Broadcast socket: Bestfive: ${JSON.stringify(data)}`);
                        callback(data);
                    });
                    break;
                case constants.EVT_SCK_TOUCHLINE:
                    self.eventEmitter.on(constants.EVT_SCK_TOUCHLINE, (data) => {
                        if (self.debug) logger.info(`Broadcast socket: Touchline: ${JSON.stringify(data)}`);
                        callback(data);
                    });
                    break;
                default:
                    break;
            };
        };
    }
}

module.exports = B2CAPI;