const WebSocketClient = require('isomorphic-ws');
const commonMethods = require('./common-methods');
const constants = require('./constants');
const { v4: UUID } = require('uuid');
const APIParser = require('./request-parser');
const zlib = require('./zlib');
const logger = require('./logger');

class SocketClient {
    constructor() {
        const self = this;
        self.client;
        self.socketReq;
        self.connectTime;
        self.webSocket;
        self.reconnectCnt;
        self.isConnected = false;
        self.id;
        self.socketState = 0;
        self.eventEmitter = commonMethods.pubsub();
        self.parser = new APIParser();
        self.IsUncompress = false;
        self.msgCodeKeyToCheck = constants.C_S_FIELD_DELIMITER + constants.C_S_TAG_MSGCODE + constants.C_S_NAMEVALUE_DELIMITER;

        self.initialize = function (socketReq) {

            try {
                self.client = new WebSocketClient(socketReq.SocketUrl);
                self.socketReq = socketReq;
                self.reconnectCnt = 0;
                self.socketState = constants.SCK_STATUS_CONNECTING;

                self.client.onopen = function (msg) {
                    self.id = UUID();
                    let sInfoMsg = `Broadcast websocket: connected.`;
                    if (self.debug) logger.info(sInfoMsg);
                    self.isConnected = true;
                    self.reconnectCnt = 0;
                    self.socketState = constants.SCK_STATUS_CONNECTED;
                    self.eventEmitter.emit(constants.EVT_SCK_OPEN, { data: { url: self.socketReq.SocketUrl, status: constants.EVT_SCK_OPEN } });
                    self.sendLogin(self.socketReq.objLogin);
                };

                self.client.onmessage = function (_event) {
                    try {
                        let fullMsg = self.defragPacket(_event.data);
                        if (fullMsg != undefined && fullMsg.length > 0) {
                            while (fullMsg.length > 0) {
                                let msg = fullMsg.pop();
                                let _response = String.fromCharCode.apply(null, msg);
                                //remove End of response char
                                if (_response.indexOf('|50=') == -1) {
                                    let intTmtrIndex = _response.indexOf(constants.C_S_CHAR0);
                                    if (intTmtrIndex != -1) {
                                        _response = _response.substr(0, intTmtrIndex);
                                    }
                                }
                                //split multi response packet with Start of response char C_S_CHAR2
                                //self.bIsAppResumeBcastReceived = true;
                                let arrData = [];
                                if (_response.indexOf('|50=') == -1) {
                                    arrData = _response.split(constants.C_S_CHAR2);
                                } else {
                                    //console.log(_response);
                                    arrData = self.parseMessage(_response); // _response.split('|64=');
                                };
                                let intDataCount = arrData.length;
                                //console.log('Split count: ',intDataCount);
                                for (let intDataCntr = 0; intDataCntr < intDataCount; intDataCntr++) {
                                    if (arrData[intDataCntr] != '') {
                                        //console.log('Split: ',arrData[intDataCntr]);
                                        self.processPacketString(arrData[intDataCntr]);
                                    };
                                };
                            };
                        };
                    } catch (error) {
                        let sErrMsg = ' WebSocket Client Message Receive Error: ' + error.message;
                        // commonMethods.writeConsoleLog(sErrMsg);
                    };
                };

                self.client.onerror = async function (msg) {
                    try {
                        self.isConnected = false;
                        self.socketState = constants.SCK_STATUS_CLOSED;
                        let sInfoMsg = `Broadcast websocket: client connection ${msg.error.toString()}`;
                        if (self.debug) logger.error(sInfoMsg);
                        self.eventEmitter.emit(constants.EVT_SCK_ERROR, { data: { url: self.socketReq.SocketUrl, status: constants.EVT_SCK_ERROR } });
                    } catch (error) {
                        let sErrMsg = `Broadcast websocket: client Error: ${error.message}`;
                        if (self.debug) logger.error(sErrMsg);
                    }
                };
                self.client.onclose = async function (msg) {
                    try {
                        self.isConnected = false;
                        self.socketState = constants.SCK_STATUS_CLOSED;
                        if (self.keepAliveId) {
                            clearTimeout(self.keepAliveId);
                        }
                        let sInfoMsg = `Broadcast websocket: client connection closed`;
                        if (self.debug) logger.error(sInfoMsg);
                        self.eventEmitter.emit(constants.EVT_SCK_CLOSE, { data: { url: self.socketReq.SocketUrl, status: constants.EVT_SCK_CLOSE } });
                    } catch (error) {
                        let sErrMsg = 'Broadcast websocket: client close Error: ' + error.message;
                        if (self.debug) logger.error(sErrMsg);
                    }
                };
                let sInfoMsg = `Connecting broadcast websocket.`;
                if (self.debug) logger.info(sInfoMsg);
            } catch (error) {
                const sInfoMsg = `Broadcast websocket: client initialize Error: ${(error.toString())}`;
                if (self.debug) logger.error(sInfoMsg);
            };
        };

        self.sendLogin = function (reqLogin) {
            try {
                const loginReq = self.parser.createLoginRequest(reqLogin);
                self.sendMessage(loginReq);
            } catch (error) {
                // console.log(error);
            };
        };

        self.sendTouchlineReq = function (reqTL) {
            try {
                const TLReq = self.parser.createTouchLineRequest(reqTL.operation, reqTL.arrScrip, true);
                if (self.isConnected) {
                    self.sendMessage(TLReq);
                } else {
                    setTimeout(() => {
                        self.sendMessage(TLReq);
                    }, 500);
                };
            } catch (error) {
                // console.log(error);
            };
        };

        self.sendBestFiveReq = function (reqB5) {
            try {
                const B5Req = self.parser.createBestFiveRequest(reqB5);
                if (self.isConnected) {
                    self.sendMessage(B5Req);
                } else {
                    setTimeout(() => {
                        self.sendMessage(B5Req);
                    }, 500);
                };
            } catch (error) {
                // console.log(error);
            };
        };

        self.sendMessage = function (_request) {
            try {
                if (self.client != null && self.client.readyState == 1) {
                    if (self.debug) logger.info('Broadcast websocket: REQ: ' + _request);
                    let dataToSend = self.fragmentData(_request);
                    self.client.send(dataToSend);
                };
            } catch (error) {
                // console.log(error);
            };
        };

        self.fragmentData = function (_requestPacket) {
            try {

                let _strHead = String.fromCharCode(2); //No compression
                _strHead = String.fromCharCode(5); //5 comprression char

                let i;
                let _data = new ArrayBuffer(_strHead.length);
                let _headerBytes = new Uint8Array(_data);

                for (i = 0; i < _strHead.length; i += 1) {
                    _headerBytes[i] = _strHead.charCodeAt(i);
                }
                let _baRequest;
                _baRequest = self.handleCompressedData(_requestPacket);

                let _length = _baRequest.length;
                _length += 4;

                let _lenLength = _length.toString().length;
                let _lengthString = '';

                for (i = 0; i < (5 - _lenLength); i++) {
                    _lengthString += '0';
                }

                _lengthString += _length.toString();

                _data = new ArrayBuffer(_lengthString.length);
                let _lenBytes = new Uint8Array(_data);

                for (i = 0; i < _lengthString.length; i += 1) {
                    _lenBytes[i] = _lengthString.charCodeAt(i);
                }

                let _baActualSend = new Uint8Array(5 + _length);
                _baActualSend.set(_lenBytes);
                _baActualSend.set(_baRequest, 5);

                let _outputStream = new Uint8Array(_headerBytes.length + _baActualSend.length);
                _outputStream.set(_headerBytes);
                _outputStream.set(_baActualSend, 1);
                return _outputStream.buffer;

            }
            catch (e) {
                ///clsGlobal.logManager.writeLog(clsGlobal.ComId, constants.C_S_LOGTYPE_APPERROR, undefined, clsGlobal.logManager.createLogPayload('clsWebSocket', 'addHTTPHeader', e.Message, undefined, e.stack, undefined, undefined));        
            }
        };

        self.handleCompressedData = function (_rawData) {
            try {
                let _data = new ArrayBuffer(_rawData.length);
                let _uint8buf = new Uint8Array(_data);
                for (let i = 0; i < _rawData.length; i += 1) {
                    _uint8buf[i] = _rawData.charCodeAt(i) & 0xFF;
                }
                //alert('CompressData');
                let _compData = zlib.compress(new Uint8Array(_data), 6);
                //alert('CompressData - After Compress' );
                return _compData;
            }
            catch (e) {
                //commonMethods.writeConsoleLog(sInfoMsg);
                //clsGlobal.logManager.writeLog(clsGlobal.ComId, clsConstants.C_S_LOGTYPE_APPERROR, undefined, clsGlobal.logManager.createLogPayload('clsWebSocket', 'handleCompressedData', e.Message, undefined, e.stack, undefined, undefined));        
            }
        };

        self.defragPacket = function (data) {

            if (data.byteLength == 0) {
                return null;
            }

            self._HeaderLength = 6;
            let dataReceived = null;
            let FullMessages = [];

            if (self.baOldData == null)
                dataReceived = new Uint8Array(data);
            else {
                dataReceived = self.appendOrCopyBuffer(self.baOldData, data);
                self.baOldData = null;
            }


            let IsDone = true;
            let MsgLength = 0;
            while (IsDone) {

                if (dataReceived.byteLength < self._HeaderLength) {
                    self.baOldData = dataReceived;
                    return FullMessages;
                }

                MsgLength = self.getMessageLength(dataReceived);
                if (MsgLength <= 0) {
                    return null;
                }

                if (dataReceived.byteLength < (MsgLength + self._HeaderLength)) {
                    self.baOldData = dataReceived;
                    return FullMessages;
                }
                else {
                    self.baOldData = null;
                }

                let CompressedMessage = new Uint8Array(dataReceived.subarray(self._HeaderLength, MsgLength + self._HeaderLength));
                let UnCompressedByteMessage = self.getUncompressedMessage(CompressedMessage);
                if (UnCompressedByteMessage == null) {
                    IsDone = false;
                    break;
                }

                FullMessages.push(UnCompressedByteMessage);

                if (dataReceived.byteLength == (MsgLength + self._HeaderLength)) {
                    IsDone = false;
                    CompressedMessage = null;
                    break;
                }

                let NewMessage = new Uint8Array(dataReceived.subarray(MsgLength + self._HeaderLength, dataReceived.byteLength));
                dataReceived = NewMessage;

            } //end of while IsDone

            dataReceived = null;
            data = null;

            return FullMessages;

        };

        self.getMessageLength = function (message) {
            if (message[0] == 5) {
                self.IsUncompress = false;
            }
            else {
                self.IsUncompress = true;
            }

            try {
                let strPacketLength = String.fromCharCode.apply(null, message.subarray(1, self._HeaderLength));
                return parseInt(strPacketLength, 10);
            }
            catch (e) {
                return 0;
            }

        };

        self.getUncompressedMessage = function (compressedMessage) {
            let uncompressedBytes = null;
            if (!self.IsUncompress) {
                let bytDecomp = null;
                bytDecomp = zlib.uncompress(compressedMessage);
                uncompressedBytes = bytDecomp;
            }
            else {
                uncompressedBytes = new Uint8Array(compressedMessage);
            }
            return uncompressedBytes;
        };

        self.appendOrCopyBuffer = function (buffer1, buffer2) {
            let tmp = new Uint8Array(buffer1.byteLength + buffer2.byteLength);
            tmp.set(new Uint8Array(buffer1), 0);
            tmp.set(new Uint8Array(buffer2), buffer1.byteLength);
            return tmp;
        };

        self.parseMessage = function (strMessage) {
            try {
                let arrMsg = [];
                if (strMessage != undefined) {
                    let index = 0;
                    while (true) {

                        let strMsgLen = (strMessage.substr(0, 6));
                        let msgLen = parseInt(strMsgLen.substr(1));
                        index += 6;
                        let strNewMsg = strMessage.substr(index, msgLen);
                        //console.log('New Message: ', strNewMsg);
                        strMessage = strMessage.substr((msgLen + 6));
                        // if(strMessage.length > 0){
                        //     console.log('Rem Message: ', strMessage);
                        // }
                        //index += msgLen;
                        index = 0;
                        //strMessage = strMessage.substr(index, msgLen);
                        arrMsg.push(strNewMsg);
                        if (strMessage == '' || strMessage == undefined) {
                            break;
                        }
                    }
                }
                return arrMsg;

            } catch (e) {
                //clsGlobal.logManager.writeLog(clsGlobal.ComId, clsConstants.C_S_LOGTYPE_APPERROR, undefined, clsGlobal.logManager.createLogPayload('clsWebSocket', 'parseMessage', e.Message, undefined, e.stack, undefined, undefined));        
                // console.log('Error parsing message: ', e.message);
                return [];
            }
        };

        self.processPacketString = function (_responsePacket) {
            try {


                let strMessageCode = '';
                let _bSendToParser = false;
                if (_responsePacket.indexOf(self.msgCodeKeyToCheck) != -1) {
                    strMessageCode = commonMethods.findValue(_responsePacket, constants.C_S_TAG_MSGCODE);
                    _bSendToParser = true;
                }

                if (_bSendToParser) {

                    _responsePacket = commonMethods.removeFieldDelimiter(_responsePacket);

                    switch (strMessageCode) {
                        case constants.C_S_MSGCODE_SOCKETLOGONRESPONSE:
                            let _objChnlLogResp = self.parser.processLogonResponse(_responsePacket, self.socketChannelId);
                            self.eventEmitter.emit(constants.EVT_SCK_LOGIN, { data: _objChnlLogResp });
                            break;
                        case constants.C_S_MSGCODE_MULTIPLE_TOUCHLINE_RESPONSE:
                            let _objMultiTLResp = self.parser.processMultiTouchLineResponse(_responsePacket);
                            self.eventEmitter.emit(constants.EVT_SCK_TOUCHLINE, { data: _objMultiTLResp });
                            break;
                        case constants.C_S_MSGCODE_BESTFIVE_RESPONSE:
                            let objB5Response = self.parser.processBestFiveResponse(_responsePacket);
                            self.eventEmitter.emit(constants.EVT_SCK_MKTDEPTH, { data: objB5Response });
                            break;
                        case constants.C_S_TAG_TRADE_EXECUTION_RANGE_RESPONSE:
                            //let objTERResponse = self.parser.processTERResponse(_responsePacket);
                            break;
                        case constants.C_S_MSGCODE_LTP_MULTIPLE_TOUCHLINE_RESPONSE:
                            //let _objLTPMultiTLResp = self.parser.processLTPMultiTouchLineResponse(_responsePacket);
                            self.eventEmitter.emit(constants.EVT_SCK_TOUCHLINE, { data: _responsePacket });
                            break;

                        case constants.C_S_MSGCODE_MTL_MULTIPLE_TOUCHLINE_RESPONSE:
                            //let _objMultiMTLResp = self.parser.processMultiMTLResponse(_responsePacket);
                            self.eventEmitter.emit(constants.EVT_SCK_TOUCHLINE, { data: _responsePacket });
                            break;
                        //PrajyotD - 20-01-2020 for MTL Touchline (e)
                        case constants.C_S_MSGCODE_APP_HEARTBEAT_RESPONSE:
                            break;
                        default:
                            break;
                    }
                }
            }
            catch (e) {
                //clsGlobal.logManager.writeLog(clsGlobal.ComId, constants.C_S_LOGTYPE_APPERROR, undefined, clsGlobal.logManager.createLogPayload('clsWebSocket', 'ProcessPacketString', e.Message, undefined, e.stack, undefined, undefined));        
            }
        };
    }
}

module.exports = SocketClient;